from typing import Dict, List, Literal

from ....action import Actions
from ....event import MoveObjectEventArguments
from ....card.support.base import UsageWithRoundRestrictionSupportBase
from ....card.support.items import ItemBase
from .....utils.class_registry import register_class
from ....match import Match
from ....modifiable_values import CostValue
from ....consts import CostLabels, IconType, ObjectPositionType
from ....struct import Cost
from .....utils.desc_registry import DescDictType


class SeedDispensary_4_3(ItemBase, UsageWithRoundRestrictionSupportBase):
    name: Literal["Seed Dispensary"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost()
    usage: int = 2
    max_usage_one_round: int = 1
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    decrease_target: int = CostLabels.SUPPORTS.value | CostLabels.EQUIPMENT.value
    decrease_threshold: int = 1

    def value_modifier_COST(
        self, value: CostValue, match: Match, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        if card label match and original cost greater than threshold,
        reduce cost.
        """
        assert value.cost.original_value is not None
        if (
            self.position.area == ObjectPositionType.SUPPORT
            and value.position.player_idx == self.position.player_idx
            and value.cost.original_value.total_dice_cost <= self.decrease_threshold
            and value.cost.label & self.decrease_target != 0
            and self.has_usage()
        ):
            # area right, player right, cost label match, cost smaller than
            # threshold, and not used this round
            if value.cost.decrease_cost(value.cost.elemental_dice_color):
                if mode == "REAL":
                    self.use()
        return value

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Match
    ) -> List[Actions]:
        if self.usage == 0:
            # if usage is zero, it run out of usage, remove it
            return list(self.check_should_remove())
        # otherwise, do parent event handler
        return super().event_handler_MOVE_OBJECT(event, match)


desc: Dict[str, DescDictType] = {
    "SUPPORT/Seed Dispensary": {
        "names": {"en-US": "Seed Dispensary", "zh-CN": "化种匣"},
        "descs": {
            "4.3": {
                "en-US": "When you play an Equipment or Support Card with an original cost of at least 1 Elemental Die: Spend 1 less Elemental Die. (Once per Round)\nUsage(s): 2",  # noqa: E501
                "zh-CN": "我方打出原本元素骰费用为1的装备或支援牌时：少花费1个元素骰。（每回合1次）\n可用次数：2",  # noqa: E501
            }
        },
        "image_path": "cardface/Assist_Prop_Seedbox.png",  # noqa: E501
        "id": 323005,
    },
}


register_class(SeedDispensary_4_3, desc)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...modifiable_values import DamageIncreaseValue

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    SkillType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    SkillTalent,
)
# Skills


class SealOfApproval(ElementalNormalAttackBase):
    name: Literal["Seal of Approval"] = "Seal of Approval"
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = ElementalNormalAttackBase.get_cost(ElementType.PYRO)


class SignedEdict(ElementalSkillBase):
    name: Literal["Signed Edict"] = "Signed Edict"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=3)
    version: Literal["3.8"] = "3.8"

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_character_status(
                "Scarlet Seal",
                {
                    "version": self.version,
                },
            )
        ]


class DoneDeal(ElementalBurstBase):
    name: Literal["Done Deal"] = "Done Deal"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_character_status("Scarlet Seal"),
            self.create_character_status("Brilliance"),
        ]


# Talents


class RightOfFinalInterpretation_3_8(SkillTalent):
    name: Literal["Right of Final Interpretation"]
    version: Literal["3.8"] = "3.8"
    character_name: Literal["Yanfei"] = "Yanfei"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO,
        elemental_dice_number=1,
        any_dice_number=2,
    )
    skill: Literal["Seal of Approval"] = "Seal of Approval"
    draw_card: bool = False

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        if not self.position.area == ObjectPositionType.CHARACTER:
            # not equipped
            return value
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not self use normal attack
            return value
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charge attack
            return value
        if value.damage_from_element_reaction:  # pragma: no cover
            # damage from element reaction
            return value
        target = match.get_object(value.target_position)
        if target.hp > 6:
            # target hp > 6
            return value
        # increase damage
        assert mode == "REAL"
        value.damage += 1
        return value


# character base


class Yanfei_3_8(CharacterBase):
    name: Literal["Yanfei"]
    version: Literal["3.8"] = "3.8"
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[SealOfApproval | SignedEdict | DoneDeal] = []
    faction: List[FactionType] = [FactionType.LIYUE]
    weapon_type: WeaponType = WeaponType.CATALYST

    def _init_skills(self) -> None:
        self.skills = [SealOfApproval(), SignedEdict(), DoneDeal()]


register_class(Yanfei_3_8 | RightOfFinalInterpretation_3_8)



from pydantic import validator

from .card.support.base import SupportBase

from .summon.base import SummonBase

from .character.character_base import CharacterBase

from .status.team_status.base import TeamStatusBase
from .struct import ObjectPosition
from ..utils import BaseModel, get_instance
from typing import Literal, List, Tuple
from ..resources.consts import CharacterIcons
from .consts import (
    DieColor,
    ELEMENT_TO_DIE_COLOR,
    ELEMENT_DEFAULT_ORDER,
    ObjectPositionType,
)
from .object_base import CardBase, ObjectBase
from .deck import Deck
from .dice import Dice


class PlayerTable(BaseModel):
    """
    Represents the player's table, which contains information about the
    player's deck information and current state of the table.

    Attributes:
        name (str): class name.
        player_name (str): The name of the player.
        player_icon (CharacterIcons): The icon of the player's character.
        player_deck_information (Deck): The description of the player's deck.

        active_character_idx (int): The index of the active character.
        has_round_ended (bool): Whether the player has declared that the
            round has ended.
        dice (List[Dice]): The list of dice on the table.
        team_status (List[Buffs]): The list of status applied to the team.
        characters (List[Characters]): The list of characters on the table.
        summons (List[Summons]): The list of summons on the table.
        supports (List[Supports]): The list of supports on the table.
        hands (List[Cards]): The list of cards in the player's hand.
        table_deck (List[Cards]): The list of cards in the table deck.
    """

    name: Literal["PlayerTable"] = "PlayerTable"
    version: Literal["0.0.1", "0.0.2", "0.0.3", "0.0.4"]

    # player information
    player_name: str = "Nahida"
    player_icon: CharacterIcons = CharacterIcons.NAHIDA
    player_deck_information: Deck = Deck()

    # table information
    player_idx: int
    active_character_idx: int = -1
    has_round_ended: bool = False
    dice: Dice = Dice()
    team_status: List[TeamStatusBase] = []
    characters: List[CharacterBase] = []
    summons: List[SummonBase] = []
    supports: List[SupportBase] = []
    hands: List[CardBase] = []
    table_deck: List[CardBase] = []
    arcane_legend: bool = True

    charge_satisfied: bool = False
    plunge_satisfied: bool = False

    # when a card is used, it will firstly be moved to using_hand, then it will
    # be moved or removed.
    using_hand: CardBase | None = None

    @validator("team_status", each_item=True, pre=True)
    def parse_team_status(cls, v):
        return get_instance(TeamStatusBase, v)

    @validator("characters", each_item=True, pre=True)
    def parse_characters(cls, v):
        return get_instance(CharacterBase, v)

    @validator("summons", each_item=True, pre=True)
    def parse_summons(cls, v):
        return get_instance(SummonBase, v)

    @validator("supports", each_item=True, pre=True)
    def parse_supports(cls, v):
        return get_instance(SupportBase, v)

    @validator("hands", "table_deck", each_item=True, pre=True)
    def parse_cards(cls, v):
        return get_instance(CardBase, v)

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.dice.position.area == ObjectPositionType.INVALID:
            # not initialized, do initialize.
            self.dice = Dice(
                position=ObjectPosition(
                    player_idx=kwargs["player_idx"],
                    area=ObjectPositionType.DICE,
                    id=-1,
                )
            )

    def dice_color_order_0_0_1(self) -> List[DieColor]:
        """
        Returns the order of dice colors.
        """
        result: List[DieColor] = [DieColor.OMNI]
        assert (
            self.active_character_idx != -1
        ), "Cannot get order when active character index not set."
        c_element = self.characters[self.active_character_idx].element
        color = ELEMENT_TO_DIE_COLOR[c_element]
        result.append(color)
        for character in self.characters:
            c_element = character.element
            color = ELEMENT_TO_DIE_COLOR[c_element]
            if color not in result:
                result.append(color)
        for element in ELEMENT_DEFAULT_ORDER:
            color = ELEMENT_TO_DIE_COLOR[element]
            if color not in result:
                result.append(color)
        return result

    def dice_color_order_0_0_2(self) -> List[DieColor]:
        """
        Returns the order of dice colors.
        Dice colors are always ELEMENT_DEFAULT_ORDER.
        """
        return [DieColor.OMNI] + [
            ELEMENT_TO_DIE_COLOR[x] for x in ELEMENT_DEFAULT_ORDER
        ]

    def dice_color_order(self) -> List[DieColor]:
        """
        Returns the order of dice colors.
        """
        if self.version == "0.0.1":
            return self.dice_color_order_0_0_1()
        return self.dice_color_order_0_0_2()

    def sort_dice(self):
        """
        Sort the dice on the table.
        """
        order = self.dice_color_order()
        self.dice.colors.sort(key=lambda x: order.index(x))

    def get_object(self, position: ObjectPosition) -> ObjectBase | None:
        """
        Get object by its position. If object not exist, return None.
        """
        if position.area == ObjectPositionType.TEAM_STATUS:
            for status in self.team_status:
                if status.id == position.id:
                    return status
            return None
        elif position.area in [
            ObjectPositionType.CHARACTER,
            ObjectPositionType.SKILL,
            ObjectPositionType.CHARACTER_STATUS,
        ]:
            character = self.characters[position.character_idx]
            return character.get_object(position)
        elif position.area == ObjectPositionType.SUMMON:
            for summon in self.summons:
                if summon.id == position.id:
                    return summon
            return None
        elif position.area == ObjectPositionType.SUPPORT:
            for support in self.supports:
                if support.id == position.id:
                    return support
            return None
        elif position.area == ObjectPositionType.HAND:
            if self.using_hand is not None and self.using_hand.id == position.id:
                return self.using_hand
            for card in self.hands:
                if card.id == position.id:
                    return card
            return None
        elif position.area == ObjectPositionType.DECK:
            for card in self.table_deck:
                if card.id == position.id:
                    return card
            return None
        else:
            raise AssertionError(f"Unknown area {position.area}.")

    def get_object_lists(self) -> List[ObjectBase]:
        """
        Get all objects in the table.
        The order of objects should follow the game rule. The rules are:
        1. objects of `self.current_player` goes first
        2. objects belongs to character goes first
            2.1. active character first, then next, next ... until all alive
                characters are included.
            2.2. for one character, order is weapon, artifact, talent, status.
            2.3. for status, order is their index in status list, i.e.
                generated time.
        3. for other objects, order is: summon, support, hand, dice, deck.
            3.1. all other objects in same region are sorted by their index in
                the list.
        """
        result: List[ObjectBase] = []
        start_character_idx = self.active_character_idx
        if start_character_idx == -1:
            start_character_idx = 0
        for i in range(len(self.characters)):
            target = (start_character_idx + i) % len(self.characters)
            if self.characters[target].is_alive:
                result += self.characters[target].get_object_lists()
        result += self.team_status
        result += self.summons
        result += self.supports
        if self.using_hand is not None:
            result.append(self.using_hand)
        result += self.hands
        result += self.table_deck

        # disable them because currently no dice and deck cards will trigger
        # events or modify values.
        # result += self.dice
        # result += self.table_deck

        return result

    def get_active_character(self) -> CharacterBase:
        """
        Returns the active character.
        """
        assert self.active_character_idx >= 0 and self.active_character_idx < len(
            self.characters
        )
        return self.characters[self.active_character_idx]

    def next_character_idx(self, current_idx: int | None = None) -> int | None:
        """
        Returns the next character index. If `current_id` is not provided, the
        active character index will be used.

        Returns:
            int: the next character ID. If there is no next character, returns
                None.
        """
        if current_idx is None:
            current_idx = self.active_character_idx
        assert self.characters[
            current_idx
        ].is_alive, "Cannot get next character when current character is not alive."
        cnum = len(self.characters)
        for i in range(1, cnum):
            target = (current_idx + i) % cnum
            if self.characters[target].is_alive:
                return target
        return None

    def previous_character_idx(self, current_idx: int | None = None) -> int | None:
        """
        Returns the previous character ID. If `current_id` is not provided, the
        active character index will be used.

        Returns:
            int: the previous character ID. If there is no previous character,
                returns None.
        """
        if current_idx is None:
            current_idx = self.active_character_idx
        assert self.characters[current_idx].is_alive, (
            "Cannot get previous character when current character is not " "alive."
        )
        cnum = len(self.characters)
        for i in range(1, cnum):
            target = (cnum + current_idx - i) % cnum
            if self.characters[target].is_alive:
                return target
        return None

    def get_charge_and_arcane_legend(self) -> Tuple[int, bool]:
        """
        Get charge of active character and arcane legend mark together.
        """
        charge = self.get_active_character().charge
        return charge, self.arcane_legend



from typing import List, Tuple, Literal

from .character.character_base import CharacterBase
from .consts import (
    ElementType,
    ElementalReactionType,
    DamageElementalType,
    DamageType,
    ELEMENT_TO_DAMAGE_TYPE,
    ObjectPositionType,
)
from .modifiable_values import DamageElementEnhanceValue, DamageIncreaseValue
from .action import CreateObjectAction
from .struct import ObjectPosition


def check_elemental_reaction(
    source: ElementType, targets: List[ElementType]
) -> Tuple[ElementalReactionType, List[ElementType], List[ElementType]]:
    """
    Check the elemental reaction based on source and targets.

    Args:
        source (ElementType): The element type of attack source.
        targets (List[ElementType]): The element application of attack targets.

    Returns:
        Tuple[ElementalReactionType, List[ElementType], List[ElementType]]:
            The elemental reaction type, the element types that are consumed,
            and the element applications updated.
    """
    assert len(targets) <= 2, f"Two element application at most, got {targets}."
    if len(targets) == 2:
        assert targets == [ElementType.CRYO, ElementType.DENDRO], (
            "Two element application must be cryo and dendro." f"Got {targets}."
        )
    assert ElementType.GEO not in targets, "Geo cannot be applied to enemies."
    assert ElementType.ANEMO not in targets, "Anemo cannot be applied to enemies."
    if source == ElementType.NONE:
        return ElementalReactionType.NONE, [], targets
    elif source == ElementType.CRYO:
        if ElementType.PYRO in targets:
            return (
                ElementalReactionType.MELT,
                [ElementType.PYRO, ElementType.CRYO],
                [],
            )
        elif ElementType.HYDRO in targets:
            return (
                ElementalReactionType.FROZEN,
                [ElementType.HYDRO, ElementType.CRYO],
                [],
            )
        elif ElementType.ELECTRO in targets:
            return (
                ElementalReactionType.SUPERCONDUCT,
                [ElementType.ELECTRO, ElementType.CRYO],
                [],
            )
        else:
            new_targets = [ElementType.CRYO]
            if ElementType.DENDRO in targets:
                new_targets.append(ElementType.DENDRO)
            return (ElementalReactionType.NONE, [], new_targets)
    elif source == ElementType.HYDRO:
        if ElementType.PYRO in targets:
            return (
                ElementalReactionType.VAPORIZE,
                [ElementType.HYDRO, ElementType.PYRO],
                [],
            )
        elif ElementType.CRYO in targets:
            return (
                ElementalReactionType.FROZEN,
                [ElementType.HYDRO, ElementType.CRYO],
                targets[1:],  # cryo must be first
            )
        elif ElementType.ELECTRO in targets:
            return (
                ElementalReactionType.ELECTROCHARGED,
                [ElementType.ELECTRO, ElementType.HYDRO],
                [],
            )
        elif ElementType.DENDRO in targets:
            return (
                ElementalReactionType.BLOOM,
                [ElementType.HYDRO, ElementType.DENDRO],
                [],
            )
        else:
            return (ElementalReactionType.NONE, [], [ElementType.HYDRO])
    elif source == ElementType.PYRO:
        if ElementType.HYDRO in targets:
            return (
                ElementalReactionType.VAPORIZE,
                [ElementType.HYDRO, ElementType.PYRO],
                [],
            )
        elif ElementType.CRYO in targets:
            return (
                ElementalReactionType.MELT,
                [ElementType.PYRO, ElementType.CRYO],
                targets[1:],  # cryo must be first
            )
        elif ElementType.ELECTRO in targets:
            return (
                ElementalReactionType.OVERLOADED,
                [ElementType.PYRO, ElementType.ELECTRO],
                [],
            )
        elif ElementType.DENDRO in targets:
            return (
                ElementalReactionType.BURNING,
                [ElementType.PYRO, ElementType.DENDRO],
                [],
            )
        else:
            return (ElementalReactionType.NONE, [], [ElementType.PYRO])
    elif source == ElementType.ELECTRO:
        if ElementType.HYDRO in targets:
            return (
                ElementalReactionType.ELECTROCHARGED,
                [ElementType.ELECTRO, ElementType.HYDRO],
                [],
            )
        elif ElementType.CRYO in targets:
            return (
                ElementalReactionType.SUPERCONDUCT,
                [ElementType.ELECTRO, ElementType.CRYO],
                targets[1:],  # cryo must be first
            )
        elif ElementType.PYRO in targets:
            return (
                ElementalReactionType.OVERLOADED,
                [ElementType.PYRO, ElementType.ELECTRO],
                [],
            )
        elif ElementType.DENDRO in targets:
            return (
                ElementalReactionType.QUICKEN,
                [ElementType.ELECTRO, ElementType.DENDRO],
                [],
            )
        else:
            return (ElementalReactionType.NONE, [], [ElementType.ELECTRO])
    elif source == ElementType.GEO:
        if ElementType.HYDRO in targets:
            return (
                ElementalReactionType.CRYSTALLIZE,
                [ElementType.GEO, ElementType.HYDRO],
                [],
            )
        elif ElementType.PYRO in targets:
            return (
                ElementalReactionType.CRYSTALLIZE,
                [ElementType.GEO, ElementType.PYRO],
                [],
            )
        elif ElementType.CRYO in targets:
            return (
                ElementalReactionType.CRYSTALLIZE,
                [ElementType.GEO, ElementType.CRYO],
                targets[1:],  # cryo must be first
            )
        elif ElementType.ELECTRO in targets:
            return (
                ElementalReactionType.CRYSTALLIZE,
                [ElementType.GEO, ElementType.ELECTRO],
                [],
            )
        else:
            return (ElementalReactionType.NONE, [], targets)
    elif source == ElementType.DENDRO:
        if ElementType.HYDRO in targets:
            return (
                ElementalReactionType.BLOOM,
                [ElementType.HYDRO, ElementType.DENDRO],
                [],
            )
        elif ElementType.PYRO in targets:
            return (
                ElementalReactionType.BURNING,
                [ElementType.PYRO, ElementType.DENDRO],
                [],
            )
        elif ElementType.ELECTRO in targets:
            return (
                ElementalReactionType.QUICKEN,
                [ElementType.ELECTRO, ElementType.DENDRO],
                [],
            )
        else:
            new_targets = [ElementType.DENDRO]
            if ElementType.CRYO in targets:
                new_targets = [ElementType.CRYO, ElementType.DENDRO]
            return (ElementalReactionType.NONE, [], new_targets)
    else:
        assert source == ElementType.ANEMO
        if ElementType.HYDRO in targets:
            return (
                ElementalReactionType.SWIRL,
                [ElementType.ANEMO, ElementType.HYDRO],
                [],
            )
        elif ElementType.PYRO in targets:
            return (
                ElementalReactionType.SWIRL,
                [ElementType.ANEMO, ElementType.PYRO],
                [],
            )
        elif ElementType.CRYO in targets:
            return (
                ElementalReactionType.SWIRL,
                [ElementType.ANEMO, ElementType.CRYO],
                targets[1:],  # cryo must be first
            )
        elif ElementType.ELECTRO in targets:
            return (
                ElementalReactionType.SWIRL,
                [ElementType.ANEMO, ElementType.ELECTRO],
                [],
            )
        else:
            return (ElementalReactionType.NONE, [], targets)


def apply_elemental_reaction(
    target_characters: List[CharacterBase],
    active_character_idx: int,
    damage: DamageElementEnhanceValue,
    reaction: ElementalReactionType,
    reacted_elements: List[ElementType],
    version: Literal["3.3", "3.4"],
) -> Tuple[
    DamageIncreaseValue, List[DamageElementEnhanceValue], CreateObjectAction | None
]:
    """
    Apply elemental reaction to DamageElementEnhanceValue, and
    generate DamageIncreaseValue.
    It returns a DamageIncreaseValue related to input damage, and a list of
    DamageElementEnhanceValue that generated by elemental reaction. If new
    objects are generated with this elemental reaction, it will also return
    the CreateObjectAction.
    """
    res: List[DamageElementEnhanceValue] = []
    damage = DamageIncreaseValue.from_element_enhance_value(
        damage, reaction, reacted_elements
    )
    target_position = damage.target_position
    create_object = elemental_reaction_side_effect(
        reaction,
        target_position.player_idx,
        target_position.character_idx,
        version=version,
    )
    if reaction == ElementalReactionType.NONE:
        return damage, res, create_object
    if damage.damage_type != DamageType.DAMAGE:
        return damage, res, create_object
    if reaction in [
        ElementalReactionType.MELT,
        ElementalReactionType.VAPORIZE,
        ElementalReactionType.OVERLOADED,
    ]:
        # +2 damage types
        damage.damage += 2
    elif reaction == ElementalReactionType.SWIRL:
        # special swirl, 1 elemental damage for other characters
        element_type = reacted_elements[1]
        assert element_type != ElementType.ANEMO, "Anemo should always be first"
        attack_target = damage.target_position.character_idx
        for cnum in range(0, len(target_characters)):
            cnum = (cnum + active_character_idx) % len(target_characters)
            if cnum == attack_target:
                continue
            c = target_characters[cnum]
            if c.is_alive:
                res.append(
                    DamageElementEnhanceValue(
                        position=damage.position,
                        target_position=ObjectPosition(
                            player_idx=damage.target_position.player_idx,
                            character_idx=cnum,
                            area=damage.target_position.area,
                            id=c.id,
                        ),
                        damage=1,
                        damage_type=DamageType.DAMAGE,
                        damage_elemental_type=ELEMENT_TO_DAMAGE_TYPE[element_type],
                        cost=damage.cost.copy(),
                        damage_from_element_reaction=True,
                    )
                )
    else:
        # all others are +1 damage types
        damage.damage += 1
        if reaction in [
            ElementalReactionType.ELECTROCHARGED,
            ElementalReactionType.SUPERCONDUCT,
        ]:
            # 1 piercing dmg for other characters
            damage_type = DamageElementalType.PIERCING
            attack_target = damage.target_position.character_idx
            for cnum in range(0, len(target_characters)):
                cnum = (cnum + active_character_idx) % len(target_characters)
                if cnum == attack_target:
                    continue
                c = target_characters[cnum]
                if c.is_alive:
                    res.append(
                        DamageElementEnhanceValue(
                            position=damage.position,
                            target_position=ObjectPosition(
                                player_idx=damage.target_position.player_idx,
                                character_idx=cnum,
                                area=damage.target_position.area,
                                id=c.id,
                            ),
                            damage=1,
                            damage_type=DamageType.DAMAGE,
                            damage_elemental_type=damage_type,
                            cost=damage.cost.copy(),
                            damage_from_element_reaction=True,
                        )
                    )
    return damage, res, create_object


def elemental_reaction_side_effect_ver_3_4(
    reaction: ElementalReactionType, player_idx: int, character_idx: int
) -> CreateObjectAction | None:
    """
    Apply side effect of elemental reaction.
    """
    if reaction == ElementalReactionType.FROZEN:
        position = ObjectPosition(
            player_idx=player_idx,
            character_idx=character_idx,
            area=ObjectPositionType.CHARACTER_STATUS,
            id=-1,
        )
        return CreateObjectAction(
            object_position=position, object_name="Frozen", object_arguments={}
        )
    elif reaction == ElementalReactionType.CRYSTALLIZE:
        position = ObjectPosition(
            player_idx=1 - player_idx,
            area=ObjectPositionType.TEAM_STATUS,
            id=-1,
        )
        return CreateObjectAction(
            object_position=position, object_name="Crystallize", object_arguments={}
        )
    elif reaction == ElementalReactionType.BURNING:
        position = ObjectPosition(
            player_idx=1 - player_idx,
            area=ObjectPositionType.SUMMON,
            id=-1,
        )
        return CreateObjectAction(
            object_position=position, object_name="Burning Flame", object_arguments={}
        )
    elif reaction == ElementalReactionType.BLOOM:
        position = ObjectPosition(
            player_idx=1 - player_idx,
            area=ObjectPositionType.TEAM_STATUS,
            id=-1,
        )
        return CreateObjectAction(
            object_position=position, object_name="Dendro Core", object_arguments={}
        )
    elif reaction == ElementalReactionType.QUICKEN:
        position = ObjectPosition(
            player_idx=1 - player_idx,
            area=ObjectPositionType.TEAM_STATUS,
            id=-1,
        )
        return CreateObjectAction(
            object_position=position,
            object_name="Catalyzing Field",
            object_arguments={},
        )
    return None


def elemental_reaction_side_effect_ver_3_3(
    reaction: ElementalReactionType, player_idx: int, character_idx: int
) -> CreateObjectAction | None:
    """
    Apply side effect of elemental reaction. In 3.3, only quicken is different.
    """
    if reaction == ElementalReactionType.QUICKEN:
        position = ObjectPosition(
            player_idx=1 - player_idx,
            area=ObjectPositionType.TEAM_STATUS,
            id=-1,
        )
        return CreateObjectAction(
            object_position=position,
            object_name="Catalyzing Field",
            object_arguments={"version": "3.3"},
        )
    return elemental_reaction_side_effect_ver_3_4(reaction, player_idx, character_idx)


def elemental_reaction_side_effect(
    reaction: ElementalReactionType,
    player_idx: int,
    character_idx: int,
    version: Literal["3.3", "3.4"] = "3.4",
) -> CreateObjectAction | None:
    if version == "3.3":
        return elemental_reaction_side_effect_ver_3_3(
            reaction, player_idx, character_idx
        )
    return elemental_reaction_side_effect_ver_3_4(reaction, player_idx, character_idx)



from typing import List, Literal

from ....utils.class_registry import register_class

from ...consts import ELEMENT_TO_DAMAGE_TYPE

from ..character_base import ElementalBurstBase, ElementalNormalAttackBase
from .mirror_maiden_3_7 import MirrorMaiden_3_7 as MM_3_7
from .mirror_maiden_3_7 import InfluxBlast as IB_3_7


class InfluxBlast(IB_3_7):
    damage: int = 3


class MirrorMaiden_3_3(MM_3_7):
    version: Literal["3.3"]
    skills: List[ElementalNormalAttackBase | InfluxBlast | ElementalBurstBase] = []

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Water Ball",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            InfluxBlast(),
            ElementalBurstBase(
                name="Rippled Reflection",
                damage=5,
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalBurstBase.get_cost(self.element, 3, 2),
            ),
        ]


register_class(MirrorMaiden_3_3)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import Actions
from ...struct import Cost

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    SkillTalent,
)


# Skills


class JadeScreen(ElementalSkillBase):
    name: Literal["Jade Screen"] = "Jade Screen"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(elemental_dice_color=DieColor.GEO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_team_status(self.name),
        ]


class Starshatter(ElementalBurstBase):
    name: Literal["Starshatter"] = "Starshatter"
    damage: int = 6
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.GEO, elemental_dice_number=3, charge=3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Check if Jade Screen exists. If so, change self.damage to 8.
        """
        status = match.player_tables[self.position.player_idx].team_status
        screen_exist: bool = False
        for s in status:
            if s.name == "Jade Screen":
                screen_exist = True
                break
        if screen_exist:
            # temporary change damage to 8
            self.damage = 8
        ret = super().get_actions(match)
        self.damage = 6
        return ret


# Talents


class StrategicReserve_3_3(SkillTalent):
    name: Literal["Strategic Reserve"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Ningguang"] = "Ningguang"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.GEO,
        elemental_dice_number=4,
    )
    skill: Literal["Jade Screen"] = "Jade Screen"


# character base


class Ningguang_3_3(CharacterBase):
    name: Literal["Ningguang"]
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.GEO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[ElementalNormalAttackBase | JadeScreen | Starshatter] = []
    faction: List[FactionType] = [FactionType.LIYUE]
    weapon_type: WeaponType = WeaponType.CATALYST

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Sparkling Scatter",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            JadeScreen(),
            Starshatter(),
        ]


register_class(Ningguang_3_3 | StrategicReserve_3_3)



from typing import Any, Dict, Literal

from .....utils.class_registry import register_class
from ....modifiable_values import CostValue
from ....consts import IconType, ObjectPositionType
from ....status.character_status.base import RoundCharacterStatus, UsageCharacterStatus
from ....struct import Cost
from ....card.event.foods import AllCharacterFoodCard
from .....utils.desc_registry import DescDictType


class FishAndChipsStatus_4_3(RoundCharacterStatus, UsageCharacterStatus):
    name: Literal["Fish and Chips"] = "Fish and Chips"
    version: Literal["4.3"] = "4.3"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        If this character use skill, decrease 1 cost.
        """
        if self.usage <= 0:  # pragma: no cover
            # no usage, not modify
            return value
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not character use skill, not modify
            return value
        # modify
        ele_cost = value.cost.elemental_dice_color
        if value.cost.decrease_cost(ele_cost):  # pragma: no branch
            # decrease success
            if mode == "REAL":
                self.usage -= 1
        return value


class FishAndChips_4_3(AllCharacterFoodCard):
    name: Literal["Fish and Chips"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(any_dice_number=2)
    can_eat_only_if_damaged: bool = False


desc: Dict[str, DescDictType] = {
    "CARD/Fish and Chips": {
        "names": {"en-US": "Fish and Chips", "zh-CN": "炸鱼薯条"},
        "descs": {
            "4.3": {
                "en-US": "During this Round, all your characters will use 1 less Elemental Die when using their next Skill.\n(A character can consume at most 1 Food per Round)",  # noqa: E501
                "zh-CN": "本回合中，所有我方角色下次使用技能时少花费1个元素骰。\n（每回合每个角色最多食用1次「料理」）",  # noqa: E501
            }
        },
        "image_path": "cardface/Event_Food_Chips.png",  # noqa: E501
        "id": 333013,
    },
    "CHARACTER_STATUS/Fish and Chips": {
        "names": {"en-US": "Fish and Chips", "zh-CN": "炸鱼薯条"},
        "descs": {
            "4.3": {
                "en-US": "During this Round, this character will use 1 less Elemental Die when using their next Skill.",  # noqa: E501
                "zh-CN": "本回合中，该角色下次使用技能时少花费1个元素骰。",
            }
        },
    },
}


register_class(FishAndChips_4_3 | FishAndChipsStatus_4_3, desc)



from typing import Any, List, Literal

from ...event import CharacterReviveEventArguments

from ....utils.class_registry import register_class

from ...summon.base import AOESummonBase

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    AOESkillBase,
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class SacredCryoPearl_3_3(AOESummonBase):
    name: Literal["Sacred Cryo Pearl"] = "Sacred Cryo Pearl"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO
    damage: int = 1
    back_damage: int = 1


# Skills


class TrailOftheQilin(ElementalSkillBase):
    name: Literal["Trail of the Qilin"] = "Trail of the Qilin"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_team_status("Ice Lotus"),
        ]


class FrostflakeArrow(ElementalNormalAttackBase, AOESkillBase):
    name: Literal["Frostflake Arrow"] = "Frostflake Arrow"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.CRYO
    back_damage: int = 2
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=5)

    use_counter: int = 0

    def event_handler_CHARACTER_REVIVE(
        self, event: CharacterReviveEventArguments, match: Any
    ) -> List[Actions]:
        """
        When self is revived, clear counter
        """
        if (
            event.action.player_idx == self.position.player_idx
            and event.action.character_idx == self.position.character_idx
        ):
            self.use_counter = 0
        return []

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack, if has talent, back damage increase
        """
        self.use_counter += 1
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if character.talent is not None:
            # if have talent, may increase damage
            if self.use_counter >= 2:
                # condition satisfied
                self.back_damage = 3
                if character.talent.version == "3.3":
                    # if is old talent
                    self.damage = 3
            else:
                self.back_damage = 2
                self.damage = 2
        return super().get_actions(match)


class CelestialShower(ElementalBurstBase, AOESkillBase):
    name: Literal["Celestial Shower"] = "Celestial Shower"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.CRYO
    back_damage: int = 1
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO, elemental_dice_number=3, charge=3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [self.create_summon("Sacred Cryo Pearl")]


# Talents


class UndividedHeart_3_7(SkillTalent):
    name: Literal["Undivided Heart"]
    version: Literal["3.7"] = "3.7"
    character_name: Literal["Ganyu"] = "Ganyu"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO,
        elemental_dice_number=5,
    )
    skill: Literal["Frostflake Arrow"] = "Frostflake Arrow"


# character base


class Ganyu_3_7(CharacterBase):
    name: Literal["Ganyu"]
    version: Literal["3.7"] = "3.7"
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase | TrailOftheQilin | FrostflakeArrow | CelestialShower
    ] = []
    faction: List[FactionType] = [FactionType.LIYUE]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Liutian Archery",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            TrailOftheQilin(),
            FrostflakeArrow(),
            CelestialShower(),
        ]


register_class(Ganyu_3_7 | UndividedHeart_3_7 | SacredCryoPearl_3_3)



from typing import Any, Literal, List

from ....utils.class_registry import register_class

from ...struct import Cost

from ...consts import (
    CostLabels,
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    ElementalReactionType,
    IconType,
    ObjectPositionType,
    PlayerActionLabels,
)

from ...action import (
    Actions,
    ChangeObjectUsageAction,
    CreateDiceAction,
    DrawCardAction,
    MakeDamageAction,
    RemoveObjectAction,
    SwitchCharacterAction,
)

from ...event import (
    CharacterDefeatedEventArguments,
    ActionEndEventArguments,
    DeclareRoundEndEventArguments,
    MakeDamageEventArguments,
    MoveObjectEventArguments,
    RoundPrepareEventArguments,
    SkillEndEventArguments,
)

from ...modifiable_values import (
    CombatActionValue,
    CostValue,
    DamageIncreaseValue,
    DamageValue,
)
from .base import RoundTeamStatus, ShieldTeamStatus, TeamStatusBase, UsageTeamStatus


class ChangingShifts_3_3(UsageTeamStatus):
    name: Literal["Changing Shifts"] = "Changing Shifts"
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.SPECIAL] = IconType.SPECIAL

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        assert self.usage > 0
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
        ):
            # not self switch character, do nothing
            return value
        if not (value.cost.label & CostLabels.SWITCH_CHARACTER.value):
            # not switch character, do nothing
            return value
        # decrease 1 any cost
        if value.cost.decrease_cost(None):
            # decrease success
            if mode == "REAL":
                self.usage -= 1
        return value

    def event_handler_SWITCH_CHARACTER(
        self, event: ActionEndEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When switch character end, check whether to remove.
        """
        return self.check_should_remove()


class IHaventLostYet_4_0(RoundTeamStatus):
    name: Literal["I Haven't Lost Yet!"] = "I Haven't Lost Yet!"
    version: Literal["4.0"] = "4.0"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.DEBUFF] = IconType.DEBUFF


class FreshWindOfFreedom_4_1(RoundTeamStatus):
    name: Literal["Fresh Wind of Freedom"] = "Fresh Wind of Freedom"
    version: Literal["4.1"] = "4.1"
    usage: int = 1
    max_usage: int = 1
    activated: bool = False
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    def event_handler_CHARACTER_DEFEATED(
        self, event: CharacterDefeatedEventArguments, match: Any
    ) -> List[Actions]:
        """
        When an enemy character is defeated, mark activated.
        """
        if self.position.player_idx != event.action.player_idx:
            # enemy defeated, mark activated.
            self.activated = True
        return []

    def value_modifier_COMBAT_ACTION(
        self, value: CombatActionValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CombatActionValue:
        """
        When enemy character defeated,
        """
        assert mode == "REAL"
        if not self.activated:
            # not activated, do nothing
            return value
        assert self.usage > 0
        assert self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
        )
        # combat action end from self, if combat action, change to quick.
        if value.do_combat_action:
            value.do_combat_action = False
            self.usage -= 1
        # else:
        #     # Mona + Kaeya may trigger
        return value

    def event_handler_ACTION_END(
        self, event: ActionEndEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When action end event, check whether to remove.
        """
        return self.check_should_remove()


class LeaveItToMe_3_3(UsageTeamStatus):
    name: Literal["Leave It to Me!"] = "Leave It to Me!"
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.SPECIAL] = IconType.SPECIAL

    def value_modifier_COMBAT_ACTION(
        self, value: CombatActionValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CombatActionValue:
        assert self.usage > 0
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
        ):
            # not self switch character, do nothing
            return value
        if value.action_label & PlayerActionLabels.SWITCH.value == 0:
            # not switch character, do nothing
            return value
        if not value.do_combat_action:
            # already quick action, do nothing
            return value
        # self switch character, change to quick action
        value.do_combat_action = False
        assert mode == "REAL"
        self.usage -= 1
        return value

    def event_handler_ACTION_END(
        self, event: ActionEndEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When action end event, check whether to remove.
        """
        return self.check_should_remove()


class EnduringRock_3_3(RoundTeamStatus):
    """
    Made Geo damage is determined by the following:
    in value_modifier_DAMAGE_INCREASE, check whether it is caused by our
    character and is Geo damage. If so, mark did_geo_attack.
    in event_handler_SKILL_END, check whether did_geo_attack is True. If so,
    trigger the following check. And regardless of the result, reset
    did_geo_attack to False.

    As DAMAGE_INCREASE with skill as source must have SKILL_END in the
    following, it will not mix other Geo damage or other character's skills.

    TODO: with Sparks 'n' Splash
    """

    name: Literal[
        "Elemental Resonance: Enduring Rock"
    ] = "Elemental Resonance: Enduring Rock"
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 1
    did_geo_attack: bool = False
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not our character use skill, do nothing
            return value
        if value.damage_elemental_type != ElementType.GEO:
            # not Geo damage, do nothing
            return value
        # our character use Geo damage skill, mark did_geo_attack
        self.did_geo_attack = True
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction | RemoveObjectAction]:
        """
        if self character use skill, and did geo attack, check whether have
        shield team status. If so, add 3 usage. Regardless of the result, reset
        did_geo_attack to False.
        """
        if not self.did_geo_attack:
            # not fit condition, do nothing
            return []
        # reset did_geo_attack
        self.did_geo_attack = False
        if (
            self.position.player_idx != event.action.position.player_idx
        ):  # pragma: no cover
            # not self character use skill, do nothing
            return []
        team_status = match.player_tables[self.position.player_idx].team_status
        for status in team_status:
            if issubclass(type(status), ShieldTeamStatus):
                # find shield status, add 3 usage
                self.usage -= 1
                return [
                    ChangeObjectUsageAction(
                        object_position=status.position,
                        change_usage=3,
                    )
                ] + self.check_should_remove()
        return list(self.check_should_remove())


class WhereIstheUnseenRazor_4_0(RoundTeamStatus):
    name: Literal["Where Is the Unseen Razor?"] = "Where Is the Unseen Razor?"
    version: Literal["4.0"] = "4.0"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.BUFF] = IconType.BUFF
    decrease_target: int = CostLabels.WEAPON.value

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        decrease weapons cost by 2
        """
        if value.position.player_idx != self.position.player_idx:
            # not self character, do nothing
            return value
        if value.cost.label & self.decrease_target == 0:
            # not weapon, do nothing
            return value
        # try decrease twice
        success = [value.cost.decrease_cost(None), value.cost.decrease_cost(None)]
        if True in success:
            # decrease success at least once
            if mode == "REAL":
                self.usage -= 1
        return value

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        As it triggers on equipping weapon,
        When move object end, check whether to remove.
        """
        return self.check_should_remove()


class SprawlingGreenery_3_3(RoundTeamStatus):
    name: Literal[
        "Elemental Resonance: Sprawling Greenery"
    ] = "Elemental Resonance: Sprawling Greenery"
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        If we trigger elemental reaction, add 2 damage.
        """
        if (
            self.position.player_idx != value.position.player_idx
            or self.position.player_idx == value.target_position.player_idx
        ):
            # not we attack or we attack ourself, do nothing
            return value
        if value.damage_type != DamageType.DAMAGE:  # pragma: no cover
            # not damage, do nothing
            return value
        if value.element_reaction == ElementalReactionType.NONE:
            # not elemental reaction, do nothing
            return value
        if self.usage <= 0:
            # no usage, do nothing
            return value
        # we trigger elemental reaction, add 2 damage
        assert mode == "REAL"
        self.usage -= 1
        value.damage += 2
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When make damage end, check whether to remove.
        """
        return self.check_should_remove()


class ReviveOnCooldown_3_7(RoundTeamStatus):
    name: Literal["Revive on cooldown"] = "Revive on cooldown"
    version: Literal["3.7"] = "3.7"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.FOOD] = IconType.FOOD


class StoneAndContracts_3_7(TeamStatusBase):
    name: Literal["Stone and Contracts"]
    version: Literal["3.7"] = "3.7"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    def event_handler_ROUND_PREPARE(
        self, event: Any, match: Any
    ) -> List[CreateDiceAction | RemoveObjectAction]:
        """
        When round prepare, create 3 omni element and remove self.
        """
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                color=DieColor.OMNI,
                number=3,
            ),
            RemoveObjectAction(
                object_position=self.position,
            ),
        ]


class AncientCourtyard_3_8(RoundTeamStatus):
    name: Literal["Ancient Courtyard"] = "Ancient Courtyard"
    version: Literal["3.8"] = "3.8"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        decrease weapons or artifact cost by 2
        """
        if value.position.player_idx != self.position.player_idx:
            # not self character, do nothing
            return value
        if (
            value.cost.label & (CostLabels.WEAPON.value | CostLabels.ARTIFACT.value)
            == 0
        ):
            # not weapon or artifact, do nothing
            return value
        # try decrease twice
        success = [value.cost.decrease_cost(None), value.cost.decrease_cost(None)]
        if True in success:
            # decrease success at least once
            if mode == "REAL":
                self.usage -= 1
        return value

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        As it triggers on equipping weapon or artifact,
        When move object end, check whether to remove.
        """
        return self.check_should_remove()


class FatuiAmbusher_3_7(UsageTeamStatus):
    name: Literal[
        "Fatui Ambusher: Cryo Cicin Mage",
        "Fatui Ambusher: Mirror Maiden",
        "Fatui Ambusher: Pyroslinger Bracer",
        "Fatui Ambusher: Electrohammer Vanguard",
    ]
    element: DamageElementalType = DamageElementalType.PIERCING
    version: Literal["3.7"] = "3.7"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    activated_this_round: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.name == "Fatui Ambusher: Cryo Cicin Mage":
            self.element = DamageElementalType.CRYO
        elif self.name == "Fatui Ambusher: Mirror Maiden":
            self.element = DamageElementalType.HYDRO
        elif self.name == "Fatui Ambusher: Pyroslinger Bracer":
            self.element = DamageElementalType.PYRO
        else:
            assert self.name == "Fatui Ambusher: Electrohammer Vanguard"
            self.element = DamageElementalType.ELECTRO

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        reset activated_this_round
        """
        self.activated_this_round = False
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        When attached character use skill, then damage itself.
        """
        if self.activated_this_round:
            # already activated, do nothing
            return []
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not character use skill, not modify
            return []
        active_character = match.player_tables[
            self.position.player_idx
        ].get_active_character()
        if self.usage > 0:  # pragma: no branch
            self.usage -= 1
            self.activated_this_round = True
            return [
                MakeDamageAction(
                    damage_value_list=[
                        DamageValue(
                            position=self.position,
                            damage_type=DamageType.DAMAGE,
                            target_position=active_character.position,
                            damage=1,
                            damage_elemental_type=self.element,
                            cost=Cost(),
                        )
                    ]
                )
            ]
        else:
            return []  # pragma: no cover


class RhythmOfTheGreatDream_3_8(UsageTeamStatus):
    name: Literal["Rhythm of the Great Dream"] = "Rhythm of the Great Dream"
    version: Literal["3.8"] = "3.8"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        decrease weapons or artifact cost by 1
        """
        if value.position.player_idx != self.position.player_idx:
            # not self character, do nothing
            return value
        if (
            value.cost.label & (CostLabels.WEAPON.value | CostLabels.ARTIFACT.value)
            == 0
        ):
            # not weapon or artifact, do nothing
            return value
        # try decrease once
        if value.cost.decrease_cost(None):
            # decrease success
            if mode == "REAL":
                self.usage -= 1
        return value

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        As it triggers on equipping weapon or artifact,
        When move object end, check whether to remove.
        """
        return self.check_should_remove()


class WhenTheCraneReturned_3_3(UsageTeamStatus):
    name: Literal["When the Crane Returned"] = "When the Crane Returned"
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 1
    decrease_usage_when_trigger: bool = True
    icon_type: Literal[IconType.SPECIAL] = IconType.SPECIAL

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[SwitchCharacterAction]:
        """
        when our character used skill, switch to next character
        """
        if self.position.player_idx != event.action.position.player_idx:
            # not our character, do nothing
            return []
        if self.usage <= 0:  # pragma: no cover
            return []
        next_idx = match.player_tables[self.position.player_idx].next_character_idx()
        if next_idx is None:
            # no next character, do nothing
            return []
        if self.decrease_usage_when_trigger:
            self.usage -= 1
        return [
            SwitchCharacterAction(
                player_idx=self.position.player_idx,
                character_idx=next_idx,
            )
        ]

    def event_handler_SWITCH_CHARACTER(
        self, event: ActionEndEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When switch character end, check whether to remove.
        """
        return self.check_should_remove()


class WindAndFreedom_4_1(WhenTheCraneReturned_3_3, RoundTeamStatus):
    name: Literal["Wind and Freedom"] = "Wind and Freedom"
    version: Literal["4.1"] = "4.1"
    decrease_usage_when_trigger: bool = False


class WindAndFreedom_3_7(FreshWindOfFreedom_4_1):
    name: Literal["Wind and Freedom"] = "Wind and Freedom"
    version: Literal["3.7"]


class Pankration_4_1(TeamStatusBase):
    name: Literal["Pankration!"] = "Pankration!"
    version: Literal["4.1"] = "4.1"
    cost: Cost = Cost()
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.SPECIAL] = IconType.SPECIAL

    def event_handler_DECLARE_ROUND_END(
        self, event: DeclareRoundEndEventArguments, match: Any
    ) -> List[DrawCardAction | RemoveObjectAction]:
        """
        When anyone declares round end, another draw 2 cards and remove self
        """
        return [
            DrawCardAction(
                player_idx=1 - event.action.player_idx,
                number=2,
                draw_if_filtered_not_enough=True,
            ),
            RemoveObjectAction(
                object_position=self.position,
            ),
        ]


class Lyresong_4_2(WhereIstheUnseenRazor_4_0):
    name: Literal["Lyresong"] = "Lyresong"
    version: Literal["4.2"] = "4.2"
    decrease_target: int = CostLabels.ARTIFACT.value


register_class(
    FreshWindOfFreedom_4_1
    | ChangingShifts_3_3
    | IHaventLostYet_4_0
    | LeaveItToMe_3_3
    | EnduringRock_3_3
    | WhereIstheUnseenRazor_4_0
    | SprawlingGreenery_3_3
    | ReviveOnCooldown_3_7
    | StoneAndContracts_3_7
    | AncientCourtyard_3_8
    | FatuiAmbusher_3_7
    | RhythmOfTheGreatDream_3_8
    | WhenTheCraneReturned_3_3
    | WindAndFreedom_4_1
    | Pankration_4_1
    | Lyresong_4_2
    | WindAndFreedom_3_7
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class


from ...action import Actions, ChangeObjectUsageAction, CreateObjectAction
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from ..character_base import (
    CreateStatusPassiveSkill,
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Skills


class VolatileSporeCloud(ElementalSkillBase):
    name: Literal["Volatile Spore Cloud"] = "Volatile Spore Cloud"
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(elemental_dice_color=DieColor.DENDRO, elemental_dice_number=3)


class FeatherSpreading(ElementalBurstBase):
    name: Literal["Feather Spreading"] = "Feather Spreading"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.DENDRO,
        elemental_dice_number=3,
        charge=2,
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Base on usage, increase damage and reset usage
        """
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        found_status: Any = None
        for status in character.status:
            if status.name == "Radical Vitality":  # pragma: no branch
                found_status = status
                break
        else:
            raise AssertionError("Radical Vitality not found")
        if found_status.usage == 0:
            # no usage, do nothing
            return super().get_actions(match)
        # increase damage
        self.damage += found_status.usage
        ret = super().get_actions(match)
        self.damage = 4
        ret.append(
            ChangeObjectUsageAction(
                object_position=found_status.position, change_usage=-found_status.usage
            )
        )
        return ret


class RadicalVitality(CreateStatusPassiveSkill):
    name: Literal["Radical Vitality"] = "Radical Vitality"
    status_name: Literal["Radical Vitality"] = "Radical Vitality"


# Talents


class ProliferatingSpores_3_3(SkillTalent):
    name: Literal["Proliferating Spores"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Jadeplume Terrorshroom"] = "Jadeplume Terrorshroom"
    cost: Cost = Cost(elemental_dice_color=DieColor.DENDRO, elemental_dice_number=3)
    skill: Literal["Volatile Spore Cloud"] = "Volatile Spore Cloud"

    def equip(self, match: Any) -> List[CreateObjectAction]:
        """
        When equip, re-create Radical Vitality with 1 more max usage
        """
        return [
            CreateObjectAction(
                object_name="Radical Vitality",
                object_position=self.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_arguments={"max_usage": 4},
            )
        ]


# character base


class JadeplumeTerrorshroom_3_3(CharacterBase):
    name: Literal["Jadeplume Terrorshroom"]
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.DENDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase
        | VolatileSporeCloud
        | FeatherSpreading
        | RadicalVitality
    ] = []
    faction: List[FactionType] = [FactionType.MONSTER]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Majestic Dance",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            VolatileSporeCloud(),
            FeatherSpreading(),
            RadicalVitality(),
        ]


register_class(JadeplumeTerrorshroom_3_3 | ProliferatingSpores_3_3)



import json
import os
from lpsim import MatchState
from tests.utils_for_test import (
    get_test_id_from_command,
    make_respond,
    set_16_omni,
    read_from_log_json,
)


def template():
    json_fname = "sayu.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    match, agent_0, agent_1 = read_from_log_json(json_path)
    # backup json
    open("111.json", "w", encoding="utf8").write(
        open(json_path, "r", encoding="utf8").read()
    )
    json_data = json.loads(open(json_path, "r", encoding="utf8").read())
    if "command_history_raw" in json_data:
        raise AssertionError("command_history_raw already exists!")
    json_data["command_history_raw"] = json_data["command_history"]
    match.config.history_level = 0
    # modify hp
    # for i in range(2):
    #     characters = match.player_tables[i].player_deck_information.characters  # noqa: E501
    #     for c in characters:
    #         c.hp = c.max_hp = 30
    # add omnipotent guide
    set_16_omni(match)
    match.start()
    match.step()
    new_commands = [[], []]
    while True:
        if match.need_respond(0):
            agent = agent_0
            nc = new_commands[0]
        elif match.need_respond(1):
            agent = agent_1
            nc = new_commands[1]
        else:
            raise AssertionError("No need respond.")
        # add tests
        enable = [
            1,  # hp
            2,  # summon
            # 3,  # support
            # 4,  # team status
            5,  # character status
            # 6,  # hands
            # 7,  # dice
            # 8,  # charge
            # 9,  # summon element
            # 10,  # character desc
            # 11,  # hand usage availability
            12,  # element application
        ]
        # 1 for hp
        hp_str = "TEST 1"
        for i in range(2):
            for c in match.player_tables[i].characters:
                hp_str += " " + str(c.hp)
        if 1 in enable:
            nc.append(hp_str)
        for tnum, table in enumerate(match.player_tables):
            # 2 for summon usage
            summon_str = f"TEST 2 p{tnum}"
            for s in table.summons:
                summon_str += f" {s.usage}"
            if 2 in enable:
                nc.append(summon_str)
            # 3 for support
            support_str = f"TEST 3 p{tnum}"
            for s in table.supports:
                support_str += f" {s.usage}"
            if 3 in enable:
                nc.append(support_str)
            # 4 for team status
            team_str = f"TEST 4 p{tnum}"
            for s in table.team_status:
                team_str += f" {s.usage}"
            if 4 in enable:
                nc.append(team_str)
            for cnum, c in enumerate(table.characters):
                # 5 for character status
                character_str = f"TEST 5 p{tnum}c{cnum}"
                for s in c.status:
                    character_str += f" {s.usage}"
                if 5 in enable:
                    nc.append(character_str)
                cdesc_str = f"TEST 10 p{tnum}c{cnum} |{c.desc}|"
                if c.is_alive and 10 in enable:
                    nc.append(cdesc_str)
                cele_str = f"TEST 12 p{tnum}c{cnum}"
                for e in c.element_application:
                    cele_str += f" {e.value}"
                if c.is_alive and 12 in enable:
                    nc.append(cele_str)
            # 6 for hands
            hands_str = f"TEST 6 p{tnum} {len(table.hands)}"
            if 6 in enable:
                nc.append(hands_str)
            # 7 for dice
            dice_str = f"TEST 7 p{tnum}"
            for c in table.dice.colors:
                dice_str += f" {c.value}"
            if 7 in enable:
                nc.append(dice_str)
            # 9 for summon element
            summon_element_str = f"TEST 9 p{tnum}"
            for summon in table.summons:
                summon_element_str += " " + summon.damage_elemental_type.value
            if 9 in enable:
                nc.append(summon_element_str)
        # 8 for character charge
        charge_str = "TEST 8"
        for i in range(2):
            for c in match.player_tables[i].characters:
                charge_str += " " + str(c.charge)
        if 8 in enable:
            nc.append(charge_str)
        # 11 for hand usage availability
        hand_usage_str = "TEST 11"
        for req in match.requests:
            if req.name == "UseCardRequest":
                hand_usage_str += " " + str(req.card_idx)
        if 11 in enable:
            nc.append(hand_usage_str)
        # filter tests
        while True:
            nc.append(agent.commands[0])
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR
    json_data["command_history"] = new_commands
    open(json_path, "w").write((json.dumps(json_data, indent=4)))


if __name__ == "__main__":
    template()



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import DeclareRoundEndAttackSummonBase

from ...modifiable_values import DamageIncreaseValue

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class BounatifulCore_4_2(DeclareRoundEndAttackSummonBase):
    name: Literal["Bountiful Core"] = "Bountiful Core"
    version: Literal["4.2"] = "4.2"
    usage: int = 1
    max_usage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.DENDRO
    damage: int = 2
    renew_type: Literal["ADD"] = "ADD"
    extra_attack_usage: int = 2


# Skills


class DanceOfHaftkarsvar(ElementalSkillBase):
    name: Literal["Dance of Haftkarsvar"] = "Dance of Haftkarsvar"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        all_hydro_dendro = True
        characters = match.player_tables[self.position.player_idx].characters
        for c in characters:
            if c.element != ElementType.HYDRO and c.element != ElementType.DENDRO:
                all_hydro_dendro = False
                break
        if all_hydro_dendro:
            # first generate team status, then attack
            return super().get_actions(match) + [
                self.create_team_status("Golden Chalice's Bounty")
            ]
        return super().get_actions(match)


class DanceOfAbzendegiDistantDreamsListeningSpring(ElementalBurstBase):
    name: Literal[
        "Dance of Abzendegi: Distant Dreams, Listening Spring"
    ] = "Dance of Abzendegi: Distant Dreams, Listening Spring"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_opposite_character_status(match, "Lingering Aeon")
        ]


# Talents


class TheStarrySkiesTheirFlowersRain_4_2(SkillTalent):
    name: Literal["The Starry Skies Their Flowers Rain"]
    version: Literal["4.2"] = "4.2"
    character_name: Literal["Nilou"] = "Nilou"
    cost: Cost = Cost(elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3)
    skill: Literal["Dance of Haftkarsvar"] = "Dance of Haftkarsvar"

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        """
        When self Bountiful Core damage, increase damage by 1
        """
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            source_area=ObjectPositionType.CHARACTER,
            target_area=ObjectPositionType.SUMMON,
        ):
            # not equipped, or source not self summon
            return value
        summon = match.get_object(value.position)
        if summon is None or summon.name != "Bountiful Core":
            # not Bountyful Core
            return value
        # damage +1
        value.damage += 1
        return value


# character base


class Nilou_4_2(CharacterBase):
    name: Literal["Nilou"]
    version: Literal["4.2"] = "4.2"
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase
        | DanceOfHaftkarsvar
        | DanceOfAbzendegiDistantDreamsListeningSpring
    ] = []
    faction: List[FactionType] = [FactionType.SUMERU]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Dance of Samser",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            DanceOfHaftkarsvar(),
            DanceOfAbzendegiDistantDreamsListeningSpring(),
        ]


register_class(Nilou_4_2 | TheStarrySkiesTheirFlowersRain_4_2 | BounatifulCore_4_2)



from typing import Any, List, Literal

from ...event import ChangeObjectUsageEventArguments, CreateObjectEventArguments

from ....utils.class_registry import register_class

from ...action import Actions, ChangeObjectUsageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    CreateStatusPassiveSkill,
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Character status. DO NOT define here, define in server/status/characor_status
# Here is just example.


# Round status, will last for several rounds and disappear
# Skills


class SecretRiteChasmicSoulfarer(ElementalSkillBase):
    name: Literal["Secret Rite: Chasmic Soulfarer"] = "Secret Rite: Chasmic Soulfarer"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        When talent equipped and level match, increase damage by 1
        """
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        talent = character.talent
        if talent is not None:
            status = (
                match.player_tables[self.position.player_idx]
                .characters[self.position.character_idx]
                .status
            )
            for s in status:
                if s.name == "Pactsworn Pathclearer":  # pragma: no branch
                    if s.usage in talent.active_levels:
                        self.damage = 4
                    break
            else:
                raise AssertionError("No Pactsworn Pathclearer status")
        ret = super().get_actions(match)
        self.damage = 3
        return ret


class SacredRiteWolfsSwiftness(ElementalBurstBase):
    name: Literal["Sacred Rite: Wolf's Swiftness"] = "Sacred Rite: Wolf's Swiftness"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=4, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        attack and increase usage
        """
        ret = super().get_actions(match)
        status = (
            match.player_tables[self.position.player_idx]
            .characters[self.position.character_idx]
            .status
        )
        for s in status:
            if s.name == "Pactsworn Pathclearer":  # pragma: no branch
                ret.append(
                    ChangeObjectUsageAction(object_position=s.position, change_usage=2)
                )
                return ret
        else:
            raise AssertionError("No Pactsworn Pathclearer status")


class LawfulEnforcer(CreateStatusPassiveSkill):
    name: Literal["Lawful Enforcer"] = "Lawful Enforcer"
    status_name: Literal["Pactsworn Pathclearer"] = "Pactsworn Pathclearer"


# Talents


class FeatherfallJudgment_3_3(SkillTalent):
    name: Literal["Featherfall Judgment"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Cyno"] = "Cyno"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO,
        elemental_dice_number=3,
    )
    skill: Literal["Secret Rite: Chasmic Soulfarer"] = "Secret Rite: Chasmic Soulfarer"
    active_levels: List[int] = [3, 5]


class FeatherfallJudgment_4_2(SkillTalent):
    name: Literal["Featherfall Judgment"]
    version: Literal["4.2"] = "4.2"
    character_name: Literal["Cyno"] = "Cyno"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO,
        elemental_dice_number=3,
    )
    skill: Literal["Secret Rite: Chasmic Soulfarer"] = "Secret Rite: Chasmic Soulfarer"
    active_levels: List[int] = [0, 2, 4, 6]


# character base


class Cyno_3_3(CharacterBase):
    name: Literal["Cyno"]
    desc: Literal["", "transform"] = ""
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase
        | SecretRiteChasmicSoulfarer
        | SacredRiteWolfsSwiftness
        | LawfulEnforcer
    ] = []
    faction: List[FactionType] = [FactionType.SUMERU]
    weapon_type: WeaponType = WeaponType.POLEARM

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Invoker's Spear",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            SecretRiteChasmicSoulfarer(),
            SacredRiteWolfsSwiftness(),
            LawfulEnforcer(),
        ]

    def _update_desc(self, match: Any):
        status = (
            match.player_tables[self.position.player_idx]
            .characters[self.position.character_idx]
            .status
        )
        for s in status:
            if s.name == "Pactsworn Pathclearer":  # pragma: no branch
                if s.usage >= 4:
                    self.desc = "transform"
                else:
                    self.desc = ""
                return
        self.desc = ""

    def event_handler_CHANGE_OBJECT_USAGE(
        self, event: ChangeObjectUsageEventArguments, match: Any
    ) -> List[Actions]:
        self._update_desc(match)
        return []

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[Actions]:
        self._update_desc(match)
        return []


register_class(Cyno_3_3 | FeatherfallJudgment_3_3 | FeatherfallJudgment_4_2)



import os
from lpsim import MatchState
from tests.utils_for_test import (
    check_hp,
    check_usage,
    get_pidx_cidx,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
    read_from_log_json,
)


def test_counter_reset_when_revive():
    match, agent_0, agent_1 = read_from_log_json(
        os.path.join(
            os.path.dirname(__file__), "jsons", "test_counter_reset_when_revive.json"
        )  # noqa: E501
    )
    match.config.history_level = 0
    # modify hp
    # for i in range(2):
    #     characters = match.player_tables[i].player_deck_information.characters  # noqa: E501
    #     for c in characters:
    #         c.hp = c.max_hp = 30
    # add omnipotent guide
    set_16_omni(match)
    match.start()
    match.step()
    new_commands = [[], []]
    while True:
        if match.need_respond(0):
            agent = agent_0
            nc = new_commands[0]
        elif match.need_respond(1):
            agent = agent_1
            nc = new_commands[1]
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            nc.append(agent.commands[0])
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:5], hps[5:]]
                check_hp(match, hps)
            elif test_id == 5:
                pidx, cidx = get_pidx_cidx(cmd)
                check_usage(match.player_tables[pidx].characters[cidx].status, cmd[3:])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    test_counter_reset_when_revive()



from typing import Dict, Literal

from ....status.team_status.event_cards import (
    WindAndFreedom_4_1 as WindAndFreedomStatus_4_1,
)

from .....utils.class_registry import register_class
from .....utils.desc_registry import DescDictType
from ....struct import Cost
from ....card.event.resonance import WindAndFreedom_4_1


class WindAndFreedomStatus_4_3(WindAndFreedomStatus_4_1):
    version: Literal["4.3"] = "4.3"


class WindAndFreedom_4_3(WindAndFreedom_4_1):
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost()


desc: Dict[str, DescDictType] = {
    "TEAM_STATUS/Wind and Freedom": {
        "descs": {
            "4.3": {
                "zh-CN": "$TEAM_STATUS/Wind and Freedom|descs|4.1|zh-CN",
                "en-US": "$TEAM_STATUS/Wind and Freedom|descs|4.1|en-US",
            }
        },
    },
    "CARD/Wind and Freedom": {
        "descs": {
            "4.3": {
                "zh-CN": "$CARD/Wind and Freedom|descs|4.1|zh-CN",
                "en-US": "$CARD/Wind and Freedom|descs|4.1|en-US",
            }
        },
    },
}


register_class(WindAndFreedomStatus_4_3 | WindAndFreedom_4_3, desc)



from typing import List, Literal

from ....utils.class_registry import register_class
from .fatui_cryo_cicin_mage_4_1 import FatuiCryoCicinMage_4_1 as FCC_4_1
from .fatui_cryo_cicin_mage_4_1 import CryoCicins_4_1 as CC_4_1
from .fatui_cryo_cicin_mage_4_1 import MistySummons as MS_4_1
from .fatui_cryo_cicin_mage_4_1 import CicinIcicle, BlizzardBranchBlossom


class CryoCicins_3_7(CC_4_1):
    version: Literal["3.7"] = "3.7"
    decrease_only_self_damage: bool = False


class MistySummons(MS_4_1):
    version: Literal["3.7"] = "3.7"


class FatuiCryoCicinMage_3_7(FCC_4_1):
    version: Literal["3.7"]
    skills: List[CicinIcicle | MistySummons | BlizzardBranchBlossom] = []

    def _init_skills(self) -> None:
        self.skills = [
            CicinIcicle(),
            MistySummons(),
            BlizzardBranchBlossom(),
        ]


register_class(FatuiCryoCicinMage_3_7 | CryoCicins_3_7)



import logging
from typing import Any, List, Tuple
from ..server.event_handler import OmnipotentGuideEventHandler_3_3
from ..server.match import Match, MatchConfig
from ..server.deck import Deck


def get_new_match(
    decks: List[Deck],
    seed: Any = None,
    rich_mode: bool = False,
    match_config: MatchConfig | None = None,
    history_level: int = 10,
    make_skill_prediction: bool = True,
    auto_step: bool = True,
) -> Tuple[Match, Any]:
    """
    Generate new match with given conditions.

    Args:
        decks: The decks of players. If its length is zero, will not set decks
            or start the match.
        seed: The random seed. It should follow the format of
            numpy.RandomState.get_state(legacy=True) or random.Random.getstate().
        rich_mode: If True, use rich mode, at round start, players is given
            16 omni dice. Mainly used in code testing.
        match_config: The config of the match. If None, use default config.
            Mainly used to set special rules, accept illegal decks, use
            recreate mode, etc.
        history_level: The level of history. Default is 10, to record important
            actions. Will have no effect when match_config is not None.
        make_skill_prediction: If True, make skill prediction. Will have no
            effect when match_config is not None.
        auto_step: If True, auto step the match once.
    Returns:
        The generated match and its initial random state.
        If generate failed or error occurred, raise error.
    """
    if seed:
        match: Match = Match(random_state=seed)
    else:
        match: Match = Match()

    if match_config:
        match.config = match_config
    else:
        match.config.history_level = history_level
        match.config.make_skill_prediction = make_skill_prediction

    if rich_mode:
        match.config.initial_dice_number = 16
        match.event_handlers.append(OmnipotentGuideEventHandler_3_3())

    random_state = match.random_state

    if len(decks) > 0:
        match.set_deck(decks)
        start_result = match.start()
        if not start_result[0]:
            raise RuntimeError(f"Match start failed. {start_result[1]}")

    if auto_step:
        if len(decks) == 0:
            logging.warning("No deck is set, match will not auto_step.")
        else:
            match._save_history()
            match.step()

    return match, random_state



"""
Team status generated by system.
"""

from typing import Any, Literal

from ....utils.class_registry import register_class
from .base import ShieldTeamStatus, UsageTeamStatus
from ...modifiable_values import DamageIncreaseValue
from ...consts import DamageElementalType, DamageType, IconType


class CatalyzingField_3_4(UsageTeamStatus):
    """
    Catalyzing field.
    """

    name: Literal["Catalyzing Field"] = "Catalyzing Field"
    version: Literal["3.4"] = "3.4"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        Increase damage for dendro or electro damages, and decrease usage.
        """
        if value.damage_type != DamageType.DAMAGE:
            # not damage, not modify
            return value
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
        ):
            # source not self, not activate
            return value
        if not self.position.check_position_valid(
            value.target_position,
            match,
            player_idx_same=False,
            target_is_active_character=True,
        ):
            # target not enemy, or target not active character, not activate
            return value
        if (
            value.damage_elemental_type
            in [
                DamageElementalType.DENDRO,
                DamageElementalType.ELECTRO,
            ]
            and self.usage > 0
        ):
            value.damage += 1
            assert mode == "REAL"
            self.usage -= 1
        return value


class CatalyzingField_3_3(CatalyzingField_3_4):
    """
    Catalyzing field.
    """

    version: Literal["3.3"]
    usage: int = 3
    max_usage: int = 3


class DendroCore_3_3(UsageTeamStatus):
    """
    Dendro core.
    """

    name: Literal["Dendro Core"] = "Dendro Core"
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        Increase damage for electro or pyro damages by 2, and decrease usage.
        """
        if value.damage_type != DamageType.DAMAGE:
            # not damage, not modify
            return value
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
        ):
            # source not self, not activate
            return value
        if not self.position.check_position_valid(
            value.target_position,
            match,
            player_idx_same=False,
            target_is_active_character=True,
        ):
            # target not enemy, or target not active character, not activate
            return value
        if (
            value.damage_elemental_type
            in [
                DamageElementalType.ELECTRO,
                DamageElementalType.PYRO,
            ]
            and self.usage > 0
        ):
            value.damage += 2
            assert mode == "REAL"
            self.usage -= 1
        return value


class Crystallize_3_3(ShieldTeamStatus):
    """
    Crystallize.
    """

    name: Literal["Crystallize"] = "Crystallize"
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 2


register_class(
    CatalyzingField_3_4 | CatalyzingField_3_3 | DendroCore_3_3 | Crystallize_3_3
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import Actions, CreateObjectAction
from ...struct import Cost, ObjectPosition

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    SkillTalent,
)


# Skills


class JumpyDumpty(ElementalSkillBase):
    name: Literal["Jumpy Dumpty"] = "Jumpy Dumpty"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        status_usage = 1
        if self.is_talent_equipped(match):
            status_usage = 2
        return super().get_actions(match) + [
            self.create_character_status(
                "Explosive Spark",
                {
                    "usage": status_usage,
                    "max_usage": status_usage,
                },
            )
        ]


class SparksNSplash(ElementalBurstBase):
    name: Literal["Sparks 'n' Splash"] = "Sparks 'n' Splash"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=3, charge=3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Create teams status on opposite, so cannot call self.create_team_status
        """
        position = ObjectPosition(
            player_idx=1 - self.position.player_idx,
            area=ObjectPositionType.TEAM_STATUS,
            id=-1,
        )
        return super().get_actions(match) + [
            CreateObjectAction(
                object_name=self.name, object_position=position, object_arguments={}
            )
        ]


# Talents


class PoundingSurprise_3_4(SkillTalent):
    name: Literal["Pounding Surprise"]
    version: Literal["3.4"] = "3.4"
    character_name: Literal["Klee"] = "Klee"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO,
        elemental_dice_number=3,
    )
    skill: Literal["Jumpy Dumpty"] = "Jumpy Dumpty"


# character base


class Klee_3_4(CharacterBase):
    name: Literal["Klee"]
    version: Literal["3.4"] = "3.4"
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[ElementalNormalAttackBase | JumpyDumpty | SparksNSplash] = []
    faction: List[FactionType] = [FactionType.MONDSTADT]
    weapon_type: WeaponType = WeaponType.CATALYST

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Kaboom!",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            JumpyDumpty(),
            SparksNSplash(),
        ]


register_class(Klee_3_4 | PoundingSurprise_3_4)



from typing import Any, Literal

from ....utils.class_registry import register_class

from ...consts import DamageElementalType, IconType, SkillType

from ...modifiable_values import DamageIncreaseValue
from .base import (
    ElementalInfusionCharacterStatus,
    RoundCharacterStatus,
    UsageCharacterStatus,
)


class Grimheart_3_8(UsageCharacterStatus):
    name: Literal["Grimheart"] = "Grimheart"
    version: Literal["3.8"] = "3.8"
    damage: int = 3
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.ATK_UP_ICE] = IconType.ATK_UP_ICE

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if self.version == "3.5":
            self.damage = 2
        else:
            assert self.version == "3.8"
            self.damage = 3

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        """
        When character uses Icetide Vortex, increase damage by self.damage
        """
        if value.damage_from_element_reaction:
            # from element reaction, do nothing
            return value
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.ELEMENTAL_SKILL
        ):
            # not self use elemental skill, do nothing
            return value
        assert self.usage > 0
        assert mode == "REAL"
        self.usage -= 1
        value.damage += self.damage
        return value


class Grimheart_3_5(Grimheart_3_8):
    version: Literal["3.5"]
    damage: int = 2


class CryoFusionAyaka_3_3(ElementalInfusionCharacterStatus, RoundCharacterStatus):
    """
    Inherit from vanilla elemental infusion status, and has talent activated
    argument. when talent activated, it will gain electro damage +1 buff,
    and usage +1.
    """

    name: Literal["Cryo Elemental Infusion"] = "Cryo Elemental Infusion"
    mark: Literal["Kamisato Ayaka"]  # used to select right status
    desc: Literal["", "ayaka_talent"] = ""
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 1

    talent_activated: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.talent_activated:
            self.desc = "ayaka_talent"

    def renew(self, new_status: "CryoFusionAyaka_3_3") -> None:
        super().renew(new_status)
        if new_status.talent_activated:
            self.talent_activated = True
            self.desc = "ayaka_talent"

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        If talent activated, increase cryo damage dealt by this character
        by 1.
        """
        if not self.talent_activated:
            # talent not activated, do nothing
            return value
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not corresponding character use skill, do nothing
            return value
        if value.damage_elemental_type != DamageElementalType.CRYO:  # pragma: no cover
            # not cryo damage, do nothing
            return value
        assert mode == "REAL"
        value.damage += 1
        return value


register_class(Grimheart_3_8 | Grimheart_3_5 | CryoFusionAyaka_3_3)



"""
Character status generated by system.
"""

from typing import Any, List, Literal

from ....utils.class_registry import register_class
from .base import RoundCharacterStatus
from ...modifiable_values import DamageIncreaseValue
from ...consts import DamageElementalType, IconType
from ...event import MakeDamageEventArguments
from ...action import Actions


class Frozen_3_3(RoundCharacterStatus):
    """
    Frozen.
    """

    name: Literal["Frozen"] = "Frozen"
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.FROZEN] = IconType.FROZEN

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        Increase damage for pyro and physical damages to self by 2, and
        decrease usage.
        """
        if not value.is_corresponding_character_receive_damage(self.position, match):
            # not this character receive damage, not modify
            return value
        if (
            value.damage_elemental_type
            in [
                DamageElementalType.PYRO,
                DamageElementalType.PHYSICAL,
            ]
            and self.usage > 0
        ):
            value.damage += 2
            assert mode == "REAL"
            self.usage -= 1
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        When damage made, check whether the status should be removed.
        Not trigger on AFTER_MAKE_DAMAGE because when damage made, run out
        of usage, but new one is generated, should remove first then generate
        new one, otherwise newly updated status will be removed.
        """
        return list(self.check_should_remove())


register_class(Frozen_3_3)



from typing import Literal, List, Any

from .....utils.class_registry import register_base_class

from ....modifiable_values import DamageIncreaseValue

from ....event import MoveObjectEventArguments, RoundPrepareEventArguments

from ....struct import ObjectPosition, Cost

from ....action import Actions, MoveObjectAction, RemoveObjectAction

from ....consts import (
    CostLabels,
    DamageElementalType,
    ObjectPositionType,
    ObjectType,
    WeaponType,
)
from ....object_base import CardBase


class WeaponBase(CardBase):
    """
    Base class of weapons.
    """

    name: str
    cost: Cost
    version: str
    weapon_type: WeaponType
    type: Literal[ObjectType.WEAPON] = ObjectType.WEAPON
    cost_label: int = (
        CostLabels.CARD.value | CostLabels.WEAPON.value | CostLabels.EQUIPMENT.value
    )

    damage_increase: int = 1  # Almost all weapons increase the damage by 1
    remove_when_used: bool = False

    def equip(self, match: Any) -> List[Actions]:
        """
        The weapon is equipped, i.e. from hand to character. Set the status
        of weapon, and if it has actions when equipped, return the actions.
        """
        return []

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # can quip on self alive characters that have same weapon type
        ret: List[ObjectPosition] = []
        for character in match.player_tables[self.position.player_idx].characters:
            if character.is_alive and character.weapon_type == self.weapon_type:
                ret.append(character.position)
        return ret

    def is_valid(self, match: Any) -> bool:
        # can only be used when in hand
        targets = self.get_targets(match)
        return self.position.area == ObjectPositionType.HAND and len(targets) != 0

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[MoveObjectAction | RemoveObjectAction]:
        """
        Act the weapon. will place it into weapon area.
        When other weapon is equipped, remove the old one.
        """
        assert target is not None
        ret: List[MoveObjectAction | RemoveObjectAction] = []
        position = target.set_id(self.id)
        character_id = target.id
        assert position.area == ObjectPositionType.CHARACTER
        assert position.player_idx == self.position.player_idx
        characters = match.player_tables[position.player_idx].characters
        for character in characters:
            if character.id == character_id:
                # check if need to remove current weapon
                if character.weapon is not None:
                    ret.append(
                        RemoveObjectAction(
                            object_position=character.weapon.position,
                        )
                    )
        ret.append(
            MoveObjectAction(
                object_position=self.position,
                target_position=position,
            )
        )
        return ret

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        """
        When this weapon is moved from hand to character, it is considered
        as equipped, and will call `self.equip`.
        """
        if (
            event.action.object_position.id == self.id
            and event.action.object_position.area == ObjectPositionType.HAND
            and event.action.target_position.area == ObjectPositionType.CHARACTER
        ):
            # this weapon equipped from hand to character
            return self.equip(match)
        return []

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        When equipped character is using skill, increase the damage.
        """
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped
            return value
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not current character using skill
            return value
        if value.damage_elemental_type == DamageElementalType.PIERCING:
            # piercing damage
            return value
        # modify damage
        value.damage += self.damage_increase
        return value


register_base_class(WeaponBase)


class RoundEffectWeaponBase(WeaponBase):
    """
    Weapons that has round effects. Refresh their usage when equipped and
    at round preparing stage.
    Instead of setting usage, set max_usage_per_round.
    """

    name: str
    cost: Cost
    version: str
    weapon_type: WeaponType
    max_usage_per_round: int

    usage: int = 0

    def equip(self, match: Any) -> List[Actions]:
        self.usage = self.max_usage_per_round
        return []

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        self.usage = self.max_usage_per_round
        return []

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        """
        When this weapon is moved from character to character, and mark as
        reset_usage, reset usage.
        """
        if (
            event.action.object_position.id == self.id
            and event.action.object_position.area == ObjectPositionType.CHARACTER
            and event.action.target_position.area == ObjectPositionType.CHARACTER
            and event.action.reset_usage
        ):
            # this weapon equipped from character to character
            self.usage = self.max_usage_per_round
        return super().event_handler_MOVE_OBJECT(event, match)



from typing import List, Literal

from ....utils.class_registry import register_class
from ...consts import DieColor
from ...struct import Cost
from .yoimiya_3_8 import Yoimiya_3_8 as Y_3_8
from .yoimiya_3_8 import RyuukinSaxifrage as RS_3_8
from .yoimiya_3_8 import FireworkFlareUp, NiwabiFireDance


class RyuukinSaxifrage(RS_3_8):
    damage: int = 4
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=4, charge=3
    )


class Yoimiya_3_4(Y_3_8):
    version: Literal["3.4"]
    skills: List[FireworkFlareUp | NiwabiFireDance | RyuukinSaxifrage] = []

    def _init_skills(self) -> None:
        self.skills = [FireworkFlareUp(), NiwabiFireDance(), RyuukinSaxifrage()]


register_class(Yoimiya_3_4)



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....consts import DamageType, ElementalReactionType, ObjectPositionType

from ....action import Actions, DrawCardAction

from ....event import ReceiveDamageEventArguments

from ....struct import Cost
from .base import RoundEffectArtifactBase


class ShadowOfTheSandKing_4_2(RoundEffectArtifactBase):
    name: Literal["Shadow of the Sand King"]
    version: Literal["4.2"] = "4.2"
    cost: Cost = Cost(same_dice_number=1)
    max_usage_per_round: int = 1

    def equip(self, match: Any) -> List[Actions]:
        return super().equip(match) + [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=1,
                draw_if_filtered_not_enough=True,
            )
        ]

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        When opponent character take elemental reaction, draw a card.
        """
        if (
            self.position.area != ObjectPositionType.CHARACTER
            or self.position.player_idx == event.final_damage.target_position.player_idx
            or event.final_damage.element_reaction == ElementalReactionType.NONE
            or event.final_damage.damage_type != DamageType.DAMAGE
            or self.usage <= 0
            or self.position.character_idx
            != match.player_tables[self.position.player_idx].active_character_idx
        ):
            # not equipped, not opponent character, or not elemental reaction,
            # or not damage, or no usage, or self not active character
            return []
        # draw card
        self.usage -= 1
        return [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=1,
                draw_if_filtered_not_enough=True,
            )
        ]


register_class(ShadowOfTheSandKing_4_2)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...status.team_status.base import ShieldTeamStatus

from ...status.character_status.base import ShieldCharacterStatus

from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageIncreaseValue

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class StoneStele_3_7(AttackerSummonBase):
    name: Literal["Stone Stele"] = "Stone Stele"
    version: Literal["3.7"] = "3.7"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.GEO
    damage: int = 1


# Skills


class DominusLapidis(ElementalSkillBase):
    name: Literal["Dominus Lapidis"] = "Dominus Lapidis"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(elemental_dice_color=DieColor.GEO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [self.create_summon("Stone Stele")]


class DominusLapidisStrikingStone(ElementalSkillBase):
    name: Literal["Dominus Lapidis: Striking Stone"] = "Dominus Lapidis: Striking Stone"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(elemental_dice_color=DieColor.GEO, elemental_dice_number=5)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_summon("Stone Stele"),
            self.create_team_status("Jade Shield"),
        ]


class PlanetBefall(ElementalBurstBase):
    name: Literal["Planet Befall"] = "Planet Befall"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.GEO, elemental_dice_number=3, charge=3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack target, and create status for target
        """
        return super().get_actions(match) + [
            self.create_opposite_character_status(match, "Petrification", {})
        ]


# Talents


class DominanceOfEarth_3_7(SkillTalent):
    name: Literal["Dominance of Earth"]
    version: Literal["3.7"] = "3.7"
    character_name: Literal["Zhongli"] = "Zhongli"
    cost: Cost = Cost(elemental_dice_color=DieColor.GEO, elemental_dice_number=5)
    skill: Literal[
        "Dominus Lapidis: Striking Stone"
    ] = "Dominus Lapidis: Striking Stone"

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["REAL", "TEST"]
    ) -> DamageIncreaseValue:
        """
        If active character has shield, or team status has shield, increase
        geo summon damage
        """
        if (
            self.position.area != ObjectPositionType.CHARACTER
            or value.position.player_idx != self.position.player_idx
            or value.position.area != ObjectPositionType.SUMMON
            or value.damage_type != DamageType.DAMAGE
            or value.damage_elemental_type != DamageElementalType.GEO
        ):
            # not equipped, or self summon, or not geo damage
            return value
        table = match.player_tables[self.position.player_idx]
        team_status = table.team_status
        have_shield = False
        team_status = table.team_status
        for status in team_status:
            if issubclass(type(status), ShieldTeamStatus):
                have_shield = True
        if not have_shield:
            # no shield team status, check character status
            active_character = table.get_active_character()
            for status in active_character.status:
                if issubclass(type(status), ShieldCharacterStatus):
                    have_shield = True
        if not have_shield:
            # no shield
            return value
        # increase damage
        assert mode == "REAL"
        value.damage += 1
        return value


# character base


class Zhongli_3_7(CharacterBase):
    name: Literal["Zhongli"]
    version: Literal["3.7"] = "3.7"
    element: ElementType = ElementType.GEO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase
        | DominusLapidis
        | DominusLapidisStrikingStone
        | PlanetBefall
    ] = []
    faction: List[FactionType] = [FactionType.LIYUE]
    weapon_type: WeaponType = WeaponType.POLEARM

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Rain of Stone",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            DominusLapidis(),
            DominusLapidisStrikingStone(),
            PlanetBefall(),
        ]


register_class(Zhongli_3_7 | DominanceOfEarth_3_7 | StoneStele_3_7)



from typing import Dict, List, Literal

from .....utils.class_registry import register_class
from ....modifiable_values import DamageValue
from ....event import SkillEndEventArguments
from ....status.character_status.base import (
    ElementalInfusionCharacterStatus,
    RoundCharacterStatus,
)
from ....action import (
    Actions,
    ChangeObjectUsageAction,
    DrawCardAction,
    MakeDamageAction,
    RemoveObjectAction,
)
from ....match import Match
from ....struct import Cost
from ....consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    IconType,
    SkillType,
    WeaponType,
)
from ....character.character_base import (
    CharacterBase,
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    SkillBase,
    SkillTalent,
)
from .....utils.desc_registry import DescDictType


class ChiselLightMirror_4_3(ElementalInfusionCharacterStatus, RoundCharacterStatus):
    name: Literal["Chisel-Light Mirror"] = "Chisel-Light Mirror"
    version: Literal["4.3"] = "4.3"
    usage: int = 2
    max_usage: int = 3

    infused_elemental_type: DamageElementalType = DamageElementalType.DENDRO
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        """
        When self normal attack, deal 1 dendro damage. If is charged attack,
        duration +1.
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same=True, character_idx_same=True
        ):
            # not self skill
            return []
        skill: SkillBase = match.get_object(event.action.position)  # type: ignore  # noqa: E501
        assert skill is not None
        if skill.skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack
            return []
        target = match.player_tables[
            1 - self.position.player_idx
        ].get_active_character()
        ret: List[MakeDamageAction | ChangeObjectUsageAction] = [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.DAMAGE,
                        target_position=target.position,
                        damage=1,
                        damage_elemental_type=DamageElementalType.DENDRO,
                        cost=Cost(),
                    )
                ]
            ),
        ]
        if (
            match.player_tables[self.position.player_idx].charge_satisfied
            and self.usage < self.max_usage
        ):
            # is charged attack and self usage not full
            ret.append(
                ChangeObjectUsageAction(object_position=self.position, change_usage=1)
            )
        return ret


class UniversalityAnElaborationOnForm(ElementalSkillBase):
    name: Literal[
        "Universality: An Elaboration on Form"
    ] = "Universality: An Elaboration on Form"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(elemental_dice_color=DieColor.DENDRO, elemental_dice_number=3)

    def get_actions(self, match: Match) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_character_status("Chisel-Light Mirror")
        ]


class ParticularFieldFettersOfPhenomena(ElementalBurstBase):
    name: Literal[
        "Particular Field: Fetters of Phenomena"
    ] = "Particular Field: Fetters of Phenomena"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.DENDRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Match) -> List[Actions]:
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        status = None
        for s in character.status:
            if s.name == "Chisel-Light Mirror":
                status = s
                break
        if status is None:
            # attack and create 3-stack status
            return super().get_actions(match) + [
                self.create_character_status("Chisel-Light Mirror", {"usage": 3})
            ]
        else:
            # attack and change status usage
            has_talent = self.is_talent_equipped(match)
            new_usage = 3 - status.usage
            if has_talent:
                new_usage = 3
            delta_usage = new_usage - status.usage
            self.damage += status.usage
            ret = super().get_actions(match)
            self.damage -= status.usage
            if new_usage > 0:
                # change usage
                ret.append(
                    ChangeObjectUsageAction(
                        object_position=status.position, change_usage=delta_usage
                    )
                )
            if new_usage == 0:
                # remove status
                ret.append(RemoveObjectAction(object_position=status.position))
            if has_talent:
                # draw 1 card
                ret.append(
                    DrawCardAction(
                        player_idx=self.position.player_idx,
                        number=1,
                        draw_if_filtered_not_enough=True,
                    )
                )
            return ret


class Structuration_4_3(SkillTalent):
    name: Literal["Structuration"]
    version: Literal["4.3"] = "4.3"
    character_name: Literal["Alhaitham"] = "Alhaitham"
    skill: Literal[
        "Particular Field: Fetters of Phenomena"
    ] = "Particular Field: Fetters of Phenomena"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.DENDRO, elemental_dice_number=3, charge=2
    )


class Alhaitham_4_3(CharacterBase):
    name: Literal["Alhaitham"]
    version: Literal["4.3"] = "4.3"
    element: ElementType = ElementType.DENDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase
        | UniversalityAnElaborationOnForm
        | ParticularFieldFettersOfPhenomena
    ] = []
    faction: List[FactionType] = [FactionType.SUMERU]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Abductive Reasoning",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            UniversalityAnElaborationOnForm(),
            ParticularFieldFettersOfPhenomena(),
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTER/Alhaitham": {
        "names": {"en-US": "Alhaitham", "zh-CN": "艾尔海森"},
        "image_path": "cardface/Char_Avatar_Alhatham.png",  # noqa: E501
        "id": 1706,
        "descs": {"4.3": {"en-US": "", "zh-CN": ""}},
    },
    "SKILL_Alhaitham_NORMAL_ATTACK/Abductive Reasoning": {
        "names": {"en-US": "Abductive Reasoning", "zh-CN": "溯因反绎法"},
        "descs": {
            "4.3": {"en-US": "Deals 2 Physical DMG.", "zh-CN": "造成2点物理伤害。"}
        },
    },
    "SKILL_Alhaitham_ELEMENTAL_SKILL/Universality: An Elaboration on Form": {
        "names": {
            "en-US": "Universality: An Elaboration on Form",
            "zh-CN": "共相·理式摹写",
        },
        "descs": {
            "4.3": {
                "en-US": "Deals 2 Dendro DMG. This character gains Chisel-Light Mirror.",  # noqa: E501
                "zh-CN": "造成2点草元素伤害，本角色附属琢光镜。",
            }
        },
    },
    "CHARACTER_STATUS/Chisel-Light Mirror": {
        "names": {"en-US": "Chisel-Light Mirror", "zh-CN": "琢光镜"},
        "descs": {
            "4.3": {
                "en-US": "Physical DMG dealt by the character is converted to Dendro DMG.\nAfter this character performs a Normal Attack: Deals 1 Dendro DMG. If this Skill is a Charged Attack, this state's Duration (Rounds) +1.\nDuration (Rounds): 2 (Can stack, max 3 stacks)",  # noqa: E501
                "zh-CN": "角色造成的物理伤害变为草元素伤害。\n角色普通攻击后：造成1点草元素伤害。如果此技能为重击，则使此状态的持续回合+1。\n持续回合：2（可叠加，最多叠加到3回合）",  # noqa: E501
            }
        },
        "image_path": "status/Alhatham_S.png",
    },
    "SKILL_Alhaitham_ELEMENTAL_BURST/Particular Field: Fetters of Phenomena": {
        "names": {
            "en-US": "Particular Field: Fetters of Phenomena",
            "zh-CN": "殊境·显像缚结",
        },
        "descs": {
            "4.3": {
                "en-US": "Deals 4 Dendro DMG, consumes Chisel-Light Mirror, with the DMG bonus based on Chisel-Light Mirror's Duration (Rounds) consumed.\nIf Chisel-Light Mirror Duration (Rounds) consumed is 0/1/2, then apply Chisel-Light Mirror with 3/2/1 Duration (Rounds) to this character.",  # noqa: E501
                "zh-CN": "造成4点草元素伤害；消耗琢光镜，此伤害提升所消耗琢光镜的持续回合值。\n如果消耗琢光镜的持续回合为0/1/2，则为角色附属持续回合为3/2/1的琢光镜。",  # noqa: E501
            }
        },
    },
    "TALENT_Alhaitham/Structuration": {
        "names": {"en-US": "Structuration", "zh-CN": "正理"},
        "descs": {
            "4.3": {
                "en-US": "Combat Action: When your active character is Alhaitham, equip this card.\nAfter Alhaitham equips this card, immediately use Particular Field: Fetters of Phenomena once.\nWhen your Alhaitham, who has this card equipped, uses Particular Field: Fetters of Phenomena, if you expended at least 1 Duration (Rounds) of Chisel-Light Mirror, attach Chisel-Light Mirror with 3 Duration (Rounds) and draw 1 card.\n(You must have Alhaitham in your deck to add this card to your deck.)",  # noqa: E501
                "zh-CN": "战斗行动：我方出战角色为艾尔海森时，装备此牌。\n艾尔海森装备此牌后，立刻使用一次殊境·显像缚结。\n装备有此牌的艾尔海森使用殊境·显像缚结时：如果消耗了持续回合至少为1的琢光镜，则总是附属持续回合为3的琢光镜，并且抓1张牌。\n（牌组中包含艾尔海森，才能加入牌组）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Constellation_Alhatham.png",  # noqa: E501
        "id": 217061,
    },
}


register_class(ChiselLightMirror_4_3 | Structuration_4_3 | Alhaitham_4_3, desc)
